from numpy import *
from sklearn.ensemble import  GradientBoostingRegressor,AdaBoostRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.tree.tree import BaseDecisionTree

def btnet(TS_data,boosting_method="GB",ntrees=100,K='all',h=1):
    '''
    Parameters
    ----------
    TS_data: list of numpy arrays
        Each array of the list contains a time-course measurement data.
        Each row of the array corresponds to a time point and each column corresponds to a gene.
        All the arrays of the list must contain the same number of genes/columns.


    boosting_method: 'GB' or 'AD', optional
        Specifies which tree-based procedure is used: either Gradient Boosting ('GB') , AdaBoost ('AB')
        default: 'GB'

    K: 'sqrt', 'all' or a positive integer, optional. only for boosting_method='GB'.
        Specifies the number of selected attributes at each node of one tree:
        either the square root of the number of candidate regulators ('sqrt'),
        the number of candidate regulators ('all'),
        or any positive integer.
        Recommend to use all when the number of genes are not too big.
        default: 'all'

    ntrees: positive integer, optional
        Specifies the number of trees grown in an ensemble.
        default: 100

    h: positive integer, optional
        Specifies the time lag.
        Each tree model that is learned predicts the expression of the target gene at time t_{k+h},
        from the expression levels of the candidate regulators and the target gene at time t_k.

    Returns
    -------
    An adjacency matrix in which the element (i,j) is the score of the edge directed from the i-th gene to the j-th gene.
    All diagonal elements are set to zero (auto-regulations are not considered).

    '''

    if not isinstance(TS_data,ndarray):
        raise ValueError('TS_data must be a list of arrays, where each row of an array corresponds to a time point and each column corresponds to a gene')

    ngenes = TS_data[0].shape[1]

    ntimepoints_min = TS_data[0].shape[0]
    for k in range(1,len(TS_data)):
        if TS_data[k].shape[0] < ntimepoints_min:
            ntimepoints_min = TS_data[k].shape[0]


    if K is not 'sqrt' and K is not 'all' and not isinstance(K,int):
        raise ValueError('input argument K must be "sqrt", "all" or a stricly positive integer')

    if isinstance(K,int) and K <= 0:
        raise ValueError('input argument K must be "sqrt", "all" or a stricly positive integer')

    if not isinstance(ntrees,int):
        raise ValueError('input argument ntrees must be a stricly positive integer')
    elif ntrees <= 0:
        raise ValueError('input argument ntrees must be a stricly positive integer')

    if not isinstance(h,int):
        raise ValueError('input argument h must be a stricly positive integer')
    elif h <= 0:
        raise ValueError('input argument h must be a stricly positive integer')
    elif h >= ntimepoints_min:
        raise ValueError('input argument h must be stricly smaller than the number of time points in the shortest time series')

    input_idx = range(ngenes)
    W = zeros((ngenes,ngenes))

    for i in range(ngenes):

        print 'Gene %d...' % (i+1)
        ris = btnet_single(TS_data,i,input_idx,K,ntrees,h,boosting_method=boosting_method)
        W[i,:] = ris
    return transpose(W)

def btnet_single(TS_data,output_idx,input_idx,K,ntrees,h,boosting_method = 'GB'):

    ngenes = TS_data[0].shape[1]
    nexp = len(TS_data)

    nsamples = 0
    for iexp in range(nexp):
        nsamples += TS_data[iexp].shape[0]

    # Add target gene to candidate regulators
    input_idx = input_idx[:]


    if output_idx in input_idx:
        print "remove ouput_idx"
        input_idx.remove(output_idx)


    ninputs = len(input_idx)

    matrix_input = zeros((nsamples-h*nexp,ninputs))
    output = zeros(nsamples-h*nexp)

    nsamples_count = 0

    for iexp in range(nexp):
        current_timeseries = TS_data[iexp]
        npoints = current_timeseries.shape[0]
        current_timeseries_input = current_timeseries[:npoints-h,input_idx]
        current_timeseries_output = current_timeseries[h:,output_idx]
        nsamples_current = current_timeseries_input.shape[0]
        matrix_input[nsamples_count:nsamples_count+nsamples_current,:] = current_timeseries_input
        output[nsamples_count:nsamples_count+nsamples_current] = current_timeseries_output
        nsamples_count += nsamples_current

    normValue = std(output)
    output = output / normValue


    print_method='Boosted Tree'
    print_K = ninputs


    if boosting_method == 'GB':
        if K=='sqrt':
            print_K = round(sqrt(ninputs))
            treeEstimator = GradientBoostingRegressor(n_estimators=ntrees,max_features="sqrt")
        elif K=='all':
            print_K = ninputs
            treeEstimator = GradientBoostingRegressor(n_estimators=ntrees,max_features="auto")
        else:
            if K < ninputs:
                print_K = K
                treeEstimator = GradientBoostingRegressor(n_estimators=ntrees,max_features=K)
            else:
                print_K = ninputs
                treeEstimator = GradientBoostingRegressor(n_estimators=ntrees,max_features="auto")

    elif boosting_method =='AB':
        treeEstimator = AdaBoostRegressor(DecisionTreeRegressor(),n_estimators=ntrees)

    else:
        raise ValueError('Wrong boosting_method is assigned')

    print "Tree method = %s, K = %d, %d trees, time lag = %d" % (print_method,print_K,ntrees,h)


    # Learn ensemble tree
    treeEstimator.fit(matrix_input,output)

    # Compute importance scores
    feature_importances = compute_feature_importances(treeEstimator)

    ris = zeros(ngenes)
    ris[input_idx] = feature_importances
    ris[output_idx] = 0
    ris = ris / sum(ris)

    return ris


def compute_feature_importances(estimator):
    if isinstance(estimator,GradientBoostingRegressor):
        total_sum = zeros((estimator.n_features,),dtype=float64)
        for i,stage in enumerate(estimator.estimators_):
            stage_sum = sum(tree.tree_.compute_feature_importances(normalize=False) for tree in stage) / len(stage)
            total_sum += stage_sum
        importances = total_sum/len(estimator.estimators_)
        return importances

    elif isinstance(estimator,AdaBoostRegressor):
        importances = [e.tree_.compute_feature_importances(normalize=False)
                       for e in estimator.estimators_]
        importances = asarray(importances)
        if importances.shape[0] == len(estimator.estimator_weights_):
            estimator_weights = estimator.estimator_weights_[:,newaxis]
        else:
            estimator_weights = estimator.estimator_weights_[estimator.estimator_weights_!=0][:,newaxis]

        importances = importances  * estimator_weights
        return sum(importances,axis=0) / len(estimator)

    elif isinstance(estimator, BaseDecisionTree):
        return estimator.tree_.compute_feature_importances(normalize=False)

    else:
        importances = [e.tree_.compute_feature_importances(normalize=False)
                       for e in estimator.estimators_]
        importances = asarray(importances)
        return sum(importances,axis=0) / len(estimator)


